# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from typing import (
    Any,
    Dict,
)

from gql import gql

from .base import StaticQuery


class EnvironmentQuery(StaticQuery):  # pylint:disable=too-few-public-methods
    """Retrieves info about the current environment."""

    query = gql(
        """
        query getEnvironment {
            environment {
                environment {
                    workers {
                        name
                        queues
                        totalProcesses
                        activeTasks
                        reservedTasks
                        queuedTasks
                    }
                }
                errors {
                    message
                }
            }
        }
    """
    )

    @staticmethod
    def _get_availability(data: Dict[str, Any]) -> str:
        """
        Calculates the queue/worker availability
        based on active tasks and total processes.

        Parameters
        ----------
        data: Dict[str, Any]
            worker or queue data
        Returns
        -------
        float
            availability
        """
        num_active = float(data["activeTasks"])
        num_total = float(data["totalProcesses"])

        result = 1 - num_active / num_total
        return f"{result*100:.1f}%"


class GetWorkerInfoQuery(EnvironmentQuery):  # pylint:disable=too-few-public-methods
    """Formats the environment info by worker."""

    def _format_response(self, response: dict, *_) -> str:

        result = (
            f'{"Worker Name":<60} '
            f'{"Queue(s)":<30} '
            f'{"Total Processes":<20} '
            f'{"Active Tasks":<20} '
            f'{"Availability":<20} '
            f'{"Reserved Tasks":<20} '
            "\n"
        )

        for worker in response["environment"]["environment"]["workers"]:
            result += (
                "\n"
                f'{worker["name"]:<60} '
                f'{", ".join(worker["queues"]):<30} '
                f'{worker["totalProcesses"]:<20} '
                f'{worker["activeTasks"]:<20} '
                f"{self._get_availability(worker):<20}"
                f'{worker["reservedTasks"]:<20}'
                "\n"
            )

        return result


class GetQueueInfoQuery(EnvironmentQuery):  # pylint:disable=too-few-public-methods
    """Formats the environment info by queue."""

    def _format_response(self, response: Dict, *_) -> str:
        """
        Formats the query output into a readable table-like manner.

        Parameters
        ----------
        response : Dict
            environment raw query result.

        Returns
        -------
        str
            formatted query output.

        Raises
        ------
        QctrlGqlException
            Any handled error that occurred on the server
        """
        result = (
            f'{"Queue":<60} '
            f'{"Workers":<30} '
            f'{"Total Processes":<20} '
            f'{"Active Tasks":<20} '
            f'{"Availability":<20} '
            f'{"Queued Tasks":<20} '
            "\n"
        )
        env_info = {
            "worker": 0,
            "totalProcesses": 0,
            "activeTasks": 0,
            "Availability": 0,
            "queuedTasks": 0,
        }
        queues = {}

        for worker in response["environment"]["environment"]["workers"]:
            # assume one worker belongs to one queue only
            queue = worker["queues"][0]
            if queue not in queues:
                queues.update({queue: env_info.copy()})
            queues[queue]["worker"] += 1
            queues[queue]["totalProcesses"] += worker["totalProcesses"]
            queues[queue]["activeTasks"] += worker["activeTasks"]
            queues[queue]["queuedTasks"] += worker["queuedTasks"]
            queues[queue]["Availability"] = self._get_availability(queues[queue])

        for queue_name, queue_info in queues.items():
            result += (
                "\n"
                f"{queue_name:<60} "
                f'{queue_info["worker"]:<30} '
                f'{queue_info["totalProcesses"]:<20} '
                f'{queue_info["activeTasks"]:<20} '
                f'{queue_info["Availability"]:<20} '
                f'{queue_info["queuedTasks"]:<20} '
                "\n"
            )

        return result



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
from typing import List

from .builders import AsyncResult
from .queries import (
    MultiRefreshActionQuery,
    StartActionQuery,
)

LOGGER = logging.getLogger(__name__)


class ParallelExecutionCollector:
    """Collects function calls and executes them in parallel."""

    def __init__(self, qctrl: "Qctrl"):
        self._qctrl = qctrl
        self._queries: List[StartActionQuery] = []
        self._async_results: List[AsyncResult] = []

    def add(self, query: StartActionQuery, async_result: AsyncResult):
        """Collect a function call as a query and the result object."""
        self._queries.append(query)
        self._async_results.append(async_result)

    def __enter__(self):
        self._qctrl.start_collection_mode(self)

    def __exit__(self, exc_type, exc_value, traceback):
        self._qctrl.stop_collection_mode()

        # dont try to call any functions upon an exception
        if isinstance(exc_value, Exception):
            return False

        if not self._queries:
            LOGGER.warning("No functions collected. Nothing to do.")
            return True

        assert len(self._queries) == len(self._async_results)

        response = [query() for query in self._queries]

        LOGGER.debug("response:%s", response)

        if len(self._queries) != len(response):
            raise RuntimeError(
                f"Unexpected responses ({len(response)}) for queries ({len(self._queries)})"
            )

        # update results with initial data
        for index, async_result in enumerate(self._async_results):
            mutation_name = self._queries[index]._mutation_name
            response_data = response[index][mutation_name]
            async_result.update(self._qctrl.gql_env, response_data)

        # build refresh query
        refresh_queries = []

        for query in self._queries:
            refresh_queries.append(query.get_refresh_query())

        refresh_query = MultiRefreshActionQuery(self._qctrl.gql_api, *refresh_queries)

        # wait for completion of all function calls
        self._qctrl.gql_env.wait_for_completion(refresh_query, *self._async_results)

        return True



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from enum import Enum

from graphql import GraphQLEnumType

from .base import BaseHandler
from .mixins import (
    DataMixin,
    PayloadMixin,
)


class EnumHandler(BaseHandler, DataMixin, PayloadMixin):
    """Handler for GraphQLEnum types."""

    _field_type = GraphQLEnumType

    @staticmethod
    def _check_valid_value(field_type, data):
        if data not in field_type.values:
            raise ValueError(f"invalid enum value: {data}")

    def _load_data(self, field_type, data):
        self._check_valid_value(field_type, data)
        return data

    def _update_data(self, field_type, obj, data):
        self._check_valid_value(field_type, data)
        return data

    def _format_payload(self, field_type, data) -> str:
        # Accept either string or enum data.
        if isinstance(data, str):
            value = data
        elif isinstance(data, Enum):
            value = data.name
        else:
            raise ValueError(f"invalid enum value: {data}")

        return value



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
import re
import sys
from collections import namedtuple
from enum import Enum
from functools import wraps
from typing import (
    Any,
    Callable,
    Optional,
    Union,
)

import commonmark
import requests
from gql.transport.exceptions import TransportServerError
from graphql import (
    GraphQLField,
    GraphQLInputField,
    GraphQLSchema,
    GraphQLType,
)
from graphql.pyutils.undefined import UndefinedType
from packaging import version
from qctrlcommons.exceptions import QctrlException
from requests import codes
from requests.exceptions import HTTPError

from qctrl.constants import UNIVERSAL_ERROR_MESSAGE

LOGGER = logging.getLogger(__name__)


class VersionError(QctrlException):
    """Raised when QCTRL client version is incompatible with the API."""


def _is_undefined(value: any) -> bool:
    """Checks if a GraphQL value is of Undefined type.

    Parameters
    ----------
    value: any


    Returns
    -------
    bool
        True if is undefined otherwise False.
    """
    return isinstance(value, UndefinedType)


def _is_deprecated(field: Union[GraphQLField, GraphQLInputField]) -> bool:
    """Checks if the field is deprecated.

    Parameters
    ----------
    field: Union[GraphQLField, GraphQLInputField]


    Returns
    -------
    bool
        True if is deprecated field, otherwise False.

    Raises
    ------
    TypeError
        invalid field.
    """

    if isinstance(field, GraphQLField):
        return field.is_deprecated

    if isinstance(field, GraphQLInputField):
        return bool(re.search("deprecated", (field.description or "").lower()))

    raise TypeError(f"invalid field: {field}")


def abstract_property(func: Callable):
    """Decorator for a property which
    should be overridden by a subclass.
    """

    @wraps(func)
    def decorator(self):
        value = func(self)

        if value is None:
            raise ValueError("abstract property value not set")

        return value

    return property(decorator)


def _clean_text(text: Optional[str]) -> str:
    if text is None:
        return ""

    return re.sub(r"\s+", " ", text).strip()


def _convert_md_to_rst(markdown_text: str) -> str:
    """Converts markdown text to rst.
    Parameters
    ----------

    markdown_text: str
        The text to be converted to rst.

    Returns
    -------
    str
        The rst formatted text
    """
    if markdown_text is None:
        return ""
    parser = commonmark.Parser()
    ast = parser.parse(markdown_text)
    return _clean_text(_parse_to_rst(ast))


def _parse_to_rst(ast_node: commonmark.node.Node) -> str:
    """Converts the markdown formatted ast node to rst text.

    Parameters
    ----------
    ast_node: commonmark.node.Node
        The ast node to be converted to rst.

    Returns
    -------
    str
        The rst formatted text.
    """

    # convert to rst
    renderer = commonmark.ReStructuredTextRenderer()
    text = renderer.render(ast_node)

    # replace double back-tick with single back-tick
    text = text.replace("``", "`")
    function_link_regex = r"(`(\S[^\$\n]+\S)`)__"
    text = re.sub(function_link_regex, r":func:`\2`", text)

    # post processing for unconverted math
    math_block_regex = r"(.. code:: math)"
    math_inline_regex = r"(`\$(.*?)\$`)"

    text = re.sub(math_block_regex, ".. math::", text)
    text = re.sub(math_inline_regex, r":math:`\2`", text)

    reference_link_regex = r"\[\^([\d.]+)\]"
    text = re.sub(reference_link_regex, r"[\1]_", text)

    return text


def check_client_version(func):
    """
    Decorator for functions and methods that may require a minimum version for
    the Q-CTRL Python package defined by the API.
    """

    def raise_exception(exc):
        """
        Raises the `VersionError` exception if the response is a 426 (Upgrade Required).
        Raises the original exception in any other situation.
        """
        # pylint: disable=misplaced-bare-raise

        if not isinstance(exc, HTTPError):
            raise

        if (
            exc.response.status_code
            != codes.UPGRADE_REQUIRED  # pylint: disable=no-member
        ):
            raise

        raise VersionError(
            "Current version of Q-CTRL Python package is not compatible with API. "
            f"Reason: {exc.response.reason}"
        ) from None

    @wraps(func)
    def _check_client_version(*args, **kwargs):
        """
        Handles any exception from the function or method to inject a `VersionError`
        when appropriate.
        """

        try:
            return func(*args, **kwargs)

        except HTTPError as exc:
            raise_exception(exc)

        except (QctrlException, TransportServerError) as exc:
            if not hasattr(exc, "__cause__"):
                raise exc

            raise_exception(exc.__cause__)

        return None

    return _check_client_version


# a namedtuple to define basic information of a package
# name : the official name of the package
# pkg_name : installed packaged name
# url : the URL for hosting the package on PyPI
# imported when locally available
_PackageInfo = namedtuple("_PackageInfo", ["name", "pkg_name", "url"])
_QCTRL_URL = "https://pypi.org/pypi/qctrl/json"
_TOOLKIT_URL = "https://pypi.org/pypi/qctrl-toolkit/json"


class PackageRegistry(Enum):
    """
    Lists Q-CTRL related packages.
    """

    QCTRL = _PackageInfo("Q-CTRL", "qctrl", _QCTRL_URL)
    QCTRLTOOLKIT = _PackageInfo("Q-CTRL Toolkit", "qctrltoolkit", _TOOLKIT_URL)

    def check_latest_version(self):
        """
        Checks the latest version of a package in PyPI and
        shows upgrade message if the current version is outdated.
        """
        latest_version = _get_latest_version(self.value.url)
        pkg = sys.modules.get(self.value.pkg_name)
        assert pkg is not None, f"{self.value.pkg_name} is not found."
        local_version = getattr(pkg, "__version__")
        if version.parse(local_version) < version.parse(latest_version):
            print(
                f"{self.value.name} package update available. Your version is {local_version}. "
                f"New version is {latest_version}."
            )


def _get_latest_version(url) -> str:
    """
    Get the latest version of Q-CTRL python package in PyPI.

    Returns
    -------
    str
        The latest version.
    """
    contents = requests.get(url).json()
    latest_version = contents["info"]["version"]
    return latest_version


def _get_mutation_result_type(schema: GraphQLSchema, mutation_name: str) -> GraphQLType:
    """Returns the GraphQLType for the given mutation.

    Parameters
    ----------
    mutation_name : str
        The name of the mutation field in the schema.


    Returns
    -------
    GraphQLType
        Result type of the mutation

    Raises
    ------
    KeyError
        invalid mutation name.
    """
    mutation_type = schema.get_type("Mutation")
    assert mutation_type

    try:
        mutation_field = mutation_type.fields[mutation_name]
    except KeyError as error:
        raise KeyError(f"unknown mutation: {mutation_name}") from error

    return mutation_field.type


def error_handler(func: Callable) -> Any:
    """
    Catch all exception and return a generic error message with
    cause of exception.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as exc:
            LOGGER.error(exc)
            raise QctrlException(UNIVERSAL_ERROR_MESSAGE) from exc

    return wrapper



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""Dynamically created classes."""
import inspect
import sys
from enum import Enum
from functools import wraps
from types import ModuleType
from typing import Callable

from qctrlcommons.exceptions import QctrlException


class BoulderModule(Enum):
    """Defines the modules in Boulder."""

    QCTRL = "qctrl"
    GRAPH = "graph"

    @classmethod
    def is_dynamic(cls, module: "BoulderModule"):
        """Check whether a module is dynamic."""
        return module is cls.QCTRL

    @classmethod
    def get_module(cls, name: str) -> "BoulderModule":
        """Get module by name."""
        module = getattr(cls, name.upper(), None)
        if module is None:
            raise QctrlException(f"module {name} not found")

        return module


def dynamic_class(submodule: str) -> Callable:
    """Decorator to register a dynamic class. The decorated function should
    return the dynamic class to be registered.

    Parameters
    ----------
    submodule : str
        The submodule name that the class should be registered under
        e.g. `types`

    Returns
    -------
    Callable
        decorator for dynamic class
    """

    def customized_decorator(func: Callable):
        @wraps(func)
        def decorator(*args, **kwargs):
            cls = func(*args, **kwargs)
            register_dynamic_class(
                cls, submodule, BoulderModule.get_module(cls.__module__)
            )
            return cls

        return decorator

    return customized_decorator


def register_dynamic_class(cls: type, submodule: str, module: BoulderModule):
    """
    Registers the class to the given dynamic
    submodule.
    """
    assert inspect.isclass(cls)

    module_name = (
        f"{module.value}.dynamic"
        if BoulderModule.is_dynamic(module)
        else f"{module.value}"
    )
    module_name += f".{submodule}"
    cls.__module__ = module_name
    cls.__package__ = module_name

    # module exists
    if module_name in sys.modules:
        module = sys.modules[module_name]

    # create new module
    else:
        module = ModuleType(module_name)
        sys.modules[module_name] = module

    setattr(module, cls.__name__, cls)


__all__ = ["dynamic_class", "register_dynamic_class"]



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
)

from gql import (
    Client,
    gql,
)
from gql.transport.exceptions import TransportQueryError
from graphql import (
    DocumentNode,
    FieldNode,
)
from qctrlcommons.exceptions import QctrlGqlException

LOGGER = logging.getLogger(__name__)


class BaseQuery(ABC):
    """Base class for wrapping a GraphQL query with custom
    validation, error handling and result formatting.
    """

    def __init__(self, client: Client):
        self._client = client

    def __call__(self, *args, disable_result_cache: bool = False, **kwargs):
        """
        Executes the GraphQL query and returns a formatted result.

        When `disable_result_cache` is passed as `True` in a `function` request,
        it will only return after the results are persisted in the database,
        instead of returning the result at the same time it is persisted.
        """
        document = self.get_query_document()
        variable_values = self._get_variable_values(*args, **kwargs)
        return self._process(
            document, variable_values, disable_result_cache=disable_result_cache
        )

    def get_query_document(self) -> DocumentNode:
        """Public function to return the query document
        to be executed. Performs type conversion to
        DocumentNode for result of _get_query_document.
        """
        document = self._get_query_document()

        if isinstance(document, str):
            document = gql(document)

        return document

    @abstractmethod
    def _get_query_document(self) -> Union[DocumentNode, str]:
        """Returns the query document to be executed. To be
        implemented by the subclass.
        """
        raise NotImplementedError

    def _process(
        self,
        document: DocumentNode,
        variable_values: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """Processes the GraphQL query. Sends the query to the server,
        handles any errors and returns the formatted result.

        Parameters
        ----------
        document: DocumentNode
            GraphQL query to be executed.
        variable_values: Optional[Dict[str, Any]]
            required variable values for the document.

        Returns
        -------
        Any
            the formatted response provided by the subclass.
        """
        response = self._execute(document, variable_values, **kwargs)
        LOGGER.debug("response: %s", response)
        self._handle_errors(document, response)
        return self._format_response(response, document)

    def _get_variable_values(  # pylint:disable=no-self-use
        self, *args, **kwargs  # pylint:disable=unused-argument
    ) -> Dict[str, Any]:
        """Converts the args and kwargs provided when calling the
        object to the variable values dict required when executing
        the document. Performs any necessary validation. To be
        overridden by the subclass. By default, an empty dictionary
        is returned.

        Returns
        -------
        Dict[str, Any]
            the variables required to execute the document.

        Raises
        ------
        QctrlGqlException
            any validation errors.
        """
        return {}

    def _execute(
        self,
        document: DocumentNode,
        variable_values: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> dict:
        """Sends a GraphQL request to the server using the document
        and variable values.

        Parameters
        ----------
        document: DocumentNode
            GraphQL query to be executed.
        variable_values: Optional[Dict[str, Any]]
            required variable values for the document.

        Returns
        -------
        dict
            the response data from the server.

        Raises
        ------
        QctrlGqlException
            any transport errors when accessing the server.
        """
        try:
            return self._client.execute(
                document, variable_values=variable_values, **kwargs
            )
        except TransportQueryError as exc:
            raise QctrlGqlException(exc.errors) from exc

    def _format_response(  # pylint:disable=no-self-use
        self, response: dict, document: DocumentNode  # pylint:disable=unused-argument
    ) -> dict:
        """Formats the query response into the expected format. Can
        be overridden by the subclass. By default, the query response
        is returned unaltered.

        Parameters
        ----------
        response: dict
            query response data returned from server.
        document: DocumentNode
            the GraphQL document that was executed.

        Returns
        -------
        dict
            by default, returns the response.

        Raises
        ------
        QctrlGqlException
            any expected error.
        """
        return response

    @staticmethod
    def _get_document_response_keys(document: DocumentNode) -> List[str]:
        """Returns the list of root-level keys for the response
        data that contain the queried data.
        """
        keys = []

        for op_def_node in document.definitions:
            for field_node in op_def_node.selection_set.selections:
                if not isinstance(field_node, FieldNode):
                    continue

                if field_node.alias:
                    keys.append(field_node.alias.value)
                else:
                    keys.append(field_node.name.value)

        return keys

    @staticmethod
    def _get_result_for_key(response: dict, key: str) -> dict:
        if key not in response:
            raise KeyError(f"expected result key `{key}` missing: {response}")

        return response[key]

    @classmethod
    def _handle_errors(cls, document: DocumentNode, response: dict):
        """Performs error handling on the GraphQL response.

        Parameters
        ----------
        document: DocumentNode
            GraphQL document sent to server
        response: dict
            corresponding response data for document

        Raises
        ------
        QctrlGqlException
            any returned errors.
        """

        # check root level errors
        root_errors = response.get("errors")

        if root_errors:
            raise QctrlGqlException(root_errors, format_to_snake=True)

        # check query level errors
        for key in cls._get_document_response_keys(document):
            query_errors = response.get(key, {}).get("errors")

            if query_errors:
                raise QctrlGqlException(response[key]["errors"], format_to_snake=True)


class StaticQuery(BaseQuery):  # pylint:disable=too-few-public-methods
    """Base for a statically defined query."""

    query: Union[DocumentNode, str] = None

    def _get_query_document(self):
        return self.query


class PatternQuery(BaseQuery):  # pylint:disable=too-few-public-methods
    """Base for a query which can be constructed
    using string formatting.
    """

    query_pattern: str = None

    def __init__(self, client: Client, **kwargs):
        super().__init__(client)
        query_str = self.query_pattern % kwargs
        LOGGER.debug("query_str: %s", query_str)
        self._query = gql(query_str)

    def _get_query_document(self):
        return self._query



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import json
from copy import deepcopy
from typing import Any

from gql import Client
from graphql import (
    DocumentNode,
    GraphQLType,
)
from qctrlcommons.exceptions import QctrlGqlException
from qctrlcommons.serializers import DataTypeDecoder

from .base import PatternQuery
from .multi import (
    MultiQuery,
    _suffix_document,
)
from .refresh_action import RefreshActionQuery


def _format_action_result(data: dict):
    """
    Decodes the `result` from the given action data, and promotes it to the top-level response.
    """
    if data["action"].get("result") is not None:
        result_data = json.loads(data["action"].pop("result"), cls=DataTypeDecoder)
        data.update(result_data)


class StartActionQuery(PatternQuery):  # pylint:disable=too-few-public-methods
    """
    Query used to start an action.

    The per-mutation result fields are not currently populated by the API, so to handle the case
    where the result is computed immediately, this query fetches the generic action `result` dict
    instead. To make this work with the logic for loading the response into data classes, we decode
    that `result` dict and promote its contents to the top-level response, which is where the
    response loading logic expects it.
    """

    query_pattern = """
        mutation startAction {
            %(mutation_name)s(input: %(payload)s) {
                action {
                    modelId
                    name
                    status
                    errors {
                        exception
                    }
                    progress
                    result
                    runtime
                }
                errors {
                    fields
                    message
                }
             }
        }
    """

    def __init__(
        self,
        client: Client,
        env: "GraphQLCustomEnvironment",
        mutation_name: str,
        input_type: GraphQLType,
        result_type: GraphQLType,
        data: Any,
    ):
        payload = env.format_payload(input_type, data)

        super().__init__(client, mutation_name=mutation_name, payload=payload)

        self._env = env
        self._result_type = result_type
        self._mutation_name = mutation_name

    def create_result_object(self, data: Any) -> Any:
        """Creates a result object to store the response data."""
        result = self._env.load_data(self._result_type, data)

        if result.errors:
            raise QctrlGqlException(result.errors)

        return result

    def get_refresh_query(self) -> RefreshActionQuery:
        """Get the corresponding refresh query."""
        return self._env.build_refresh_query(self._result_type)

    def _format_response(self, response, *_) -> dict:
        _format_action_result(response[self._mutation_name])
        return response


class MultiStartActionQuery(MultiQuery):
    """Starts multiple actions."""

    name = "startActions"

    def _transform_sub_query_document(
        self, index: int, document: DocumentNode
    ) -> DocumentNode:
        document = deepcopy(document)
        _suffix_document(document, str(index + 1))
        return document

    def _format_response(self, response, *args):
        response = super()._format_response(response, *args)

        # for each response, format the data for updating
        for sub_response in response:
            _format_action_result(sub_response)

        return response



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
"""
Functionality related to the Q-CTRL Python Toolkit.

The toolkit objects are re-imported here to allow all access to the objects to happen directly from
the `qctrl` package.
"""

# pylint: disable=unused-import

from qctrltoolkit import (
    FUNCTIONS,
    NODES,
    TOOLKIT_ATTR,
    TOOLKIT_DOC_CONFIG,
    TOOLKIT_MAIN_DOC,
    Namespace,
    forge_toolkit,
)

__doc__ = TOOLKIT_MAIN_DOC



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring,too-many-ancestors
from typing import (
    Any,
    Dict,
)

from qctrl.builders.graphql_utils.handlers import NamedInputObjectHandler
from qctrl.builders.type_registry import TypeOverrideMixin


class ComplexNumberInputHandler(NamedInputObjectHandler, TypeOverrideMixin):
    """Custom handler for ComplexNumberInput."""

    _type_hint = complex
    _input_object_name = "ComplexNumberInput"

    def get_input_object_values(self, data: Dict[str, Any]):
        if isinstance(data, complex):
            real = data.real
            imag = data.imag

        elif isinstance(data, (int, float)):
            real = data
            imag = 0

        elif isinstance(data, dict):
            real = data.get("real", 0)
            imag = data.get("imag", 0)

        else:
            try:
                real = float(data)
            except (TypeError, ValueError) as error:
                raise TypeError(f"unknown type: {type(data)}, ({data})") from error
            else:
                imag = 0

        return {"real": real, "imag": imag}



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
from typing import (
    Any,
    Dict,
    Optional,
)

from graphql import (
    GraphQLObjectType,
    GraphQLType,
    is_object_type,
)

from qctrl.utils import abstract_property

from .base import BaseHandler
from .mixins import (
    DataMixin,
    SelectionMixin,
)

LOGGER = logging.getLogger(__name__)


class ObjectHandler(BaseHandler, SelectionMixin, DataMixin):
    """Handler for GraphQLObject types."""

    _field_type = GraphQLObjectType

    def _get_selection(self, field_type, overrides=None):
        fields = []

        for field_name, field in field_type.fields.items():

            nested_overrides = self._env.get_nested_selection_override(
                overrides, field_name
            )

            if isinstance(nested_overrides, str):
                nested_fields = nested_overrides
            else:
                nested_fields = self._env.get_selection(field.type, nested_overrides)

            if nested_fields:
                fields.append(f"{field_name} {nested_fields}")
            else:
                fields.append(field_name)

        return self._env.format_query_selection_fields(fields)

    def load_data(self, field_type: GraphQLType, data: Optional[Dict[str, Any]]) -> Any:
        """Public method for loading the given data according to
        the field type. If the data is None, no formatting is
        performed.
        """
        if data is None and not is_object_type(field_type):
            return data

        return self._load_data(field_type, data)

    def _get_nested_data(self, field_name: str, data: Dict[str, Any]):
        if field_name in data:
            return data[field_name]
        if self._env.field_to_attr(field_name) in data:
            return data[self._env.field_to_attr(field_name)]
        return None

    def _load_data(self, field_type, data):
        cls = self._env.type_registry.get_type(field_type.name)
        kwargs = {}

        for field_name, field in field_type.fields.items():

            if data is not None:
                nested_data = self._get_nested_data(field_name, data)
                if nested_data is None:
                    continue
            else:
                nested_data = None

            nested_value = self._env.load_data(field.type, nested_data)

            if nested_value is not None:
                kwargs[self._env.field_to_attr(field_name)] = nested_value

        return cls(**kwargs)

    def _update_data(self, field_type, obj, data):

        for field_name, field in field_type.fields.items():

            nested_data = self._get_nested_data(field_name, data)

            if nested_data is None:
                continue

            attr = self._env.field_to_attr(field_name)

            current_attr_value = getattr(obj, attr, None)
            if current_attr_value:
                nested_value = self._env.update_data(
                    field.type, current_attr_value, nested_data
                )
            else:
                nested_value = self._env.load_data(field.type, nested_data)

            if nested_value is not None:
                setattr(obj, attr, nested_value)

        return obj


class NamedObjectHandler(ObjectHandler):
    """Handler for overriding specific GraphQLObject types
    by name.
    """

    _object_name = None

    @abstract_property
    def object_name(self):  # pylint:disable=missing-function-docstring
        return self._object_name

    def can_handle_get_selection(self, field_type):
        return (
            super().can_handle_get_selection(field_type)
            and field_type.name  # pylint:disable=comparison-with-callable
            == self.object_name
        )

    def can_handle_data(self, field_type):
        return (
            super().can_handle_data(field_type)
            and field_type.name  # pylint:disable=comparison-with-callable
            == self.object_name
        )

    def process_loaded_data(  # pylint:disable=no-self-use
        self, data: Dict[str, Any]
    ) -> Any:
        """Override to perform any custom processing on
        loaded data. The data provided will be a dictionary
        of each sub field and its value. By default, no
        processing is done on the data.
        """
        return data

    def load_data(self, field_type: GraphQLType, data: Optional[Dict[str, Any]]) -> Any:
        """Public method for loading the given data according to
        the field type. If the data is None, no formatting is
        performed.
        """
        if data is None:
            return data

        return self._load_data(field_type, data)

    def _load_data(self, field_type, data) -> dict:

        result = {}

        for field_name, field in field_type.fields.items():

            try:
                nested_data = data[field_name]
            except KeyError:
                continue

            nested_value = self._env.load_data(field.type, nested_data)

            if nested_value is not None:
                result[field_name] = nested_value

        return self.process_loaded_data(result)

    def _update_data(self, field_type, obj, data):
        return self.load_data(field_type, data)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import json
from copy import deepcopy
from typing import Union

from gql import Client
from graphql import (
    DocumentNode,
    GraphQLNonNull,
    GraphQLType,
)
from qctrlcommons.serializers import DataTypeDecoder

from .base import PatternQuery
from .multi import (
    MultiQuery,
    _suffix_document,
)


def _format_action_result(data: dict):
    """Prepares data for updating a Result object. If the
    action result is included in the data, it is decoded
    and promoted to the top level of the data."""

    # if result data included in response
    if data["action"].get("result") is not None:

        # decode result data
        result_data = json.loads(data["action"].pop("result"), cls=DataTypeDecoder)

        # move to top level
        data.update(result_data)


class RefreshActionQuery(PatternQuery):  # pylint:disable=too-few-public-methods
    """Query used to check action completion."""

    query_pattern = """
        query refreshAction($modelId: String!) {
            coreAction(modelId: $modelId) {
                coreAction {
                    ... on %(field_type)s {
                        action {
                            modelId
                            name
                            status
                            errors {
                                exception
                            }
                            progress
                            result
                            runtime
                        }
                        errors {
                            fields
                            message
                        }
                    }
                }
                errors {
                    message
                    fields
                }
            }
        }
    """

    def __init__(self, client: Client, field_type: GraphQLType):

        if isinstance(field_type, GraphQLNonNull):
            field_type = field_type.of_type

        self.field_type = field_type
        super().__init__(client, field_type=field_type.name)

    def _get_variable_values(
        self, action_id: Union[str, int]
    ):  # pylint:disable=arguments-differ
        return {"modelId": str(action_id)}

    def _format_response(self, response, *_) -> dict:
        # retrieve and format the update data
        data = response["coreAction"]["coreAction"]
        _format_action_result(data)
        return data


class MultiRefreshActionQuery(MultiQuery):
    """Check completion for multiple actions."""

    name = "refreshActions"

    def _transform_sub_query_document(
        self, index: int, document: DocumentNode
    ) -> DocumentNode:
        document = deepcopy(document)
        _suffix_document(document, str(index + 1))
        return document

    def _get_variable_values(
        self, *action_ids: Union[str, int]
    ):  # pylint:disable=arguments-differ
        values = {}

        for index, action_id in enumerate(action_ids):
            key = f"modelId_{index+1}"
            values.update({key: action_id})

        return values

    def _format_response(self, response, *args):
        response = super()._format_response(response, *args)

        # for each response, format the data for updating
        for index, sub_response in enumerate(response):
            data = sub_response["coreAction"]
            _format_action_result(data)
            response[index] = data

        return response



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from typing import (
    Any,
    Dict,
)

from graphql import GraphQLInputObjectType

from qctrl.utils import abstract_property

from .base import BaseHandler
from .mixins import PayloadMixin


class InputObjectHandler(BaseHandler, PayloadMixin):
    """Handler for GraphQLInputObject types."""

    _field_type = GraphQLInputObjectType

    def _format_payload(self, field_type, data):
        values = {}

        for field_name, field in field_type.fields.items():
            nested_value = self._env.get_nested_value(data, field_name)

            if nested_value is not None:
                values[field_name] = self._env.format_payload(field.type, nested_value)

        return self._env.format_query_payload_dict(values)


class NamedInputObjectHandler(InputObjectHandler):
    """Handler for overriding specific GraphQLInputObject
    types by name.
    """

    _input_object_name = None

    @abstract_property
    def input_object_name(self):  # pylint:disable=missing-function-docstring
        return self._input_object_name

    def get_input_object_values(self, data: Any) -> Dict[str, Any]:
        """Given some object, return a dictionary
        of the required fields and their values.
        """
        raise NotImplementedError

    def _format_payload(self, field_type, data):
        values = self.get_input_object_values(data)

        for field_name, value in values.items():
            try:
                field = field_type.fields[field_name]
            except KeyError as error:
                raise KeyError(f"unknown field: {field_name}") from error

            values[field_name] = self._env.format_payload(field.type, value)

        return self._env.format_query_payload_dict(values)

    def can_handle_format_payload(self, field_type):
        return (
            super().can_handle_format_payload(field_type)
            and field_type.name  # pylint:disable=comparison-with-callable
            == self.input_object_name
        )



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from typing import (
    Any,
    Dict,
)

import numpy as np
from graphql import GraphQLType
from inflection import underscore
from qctrlcommons.serializers import read_numpy_array

from qctrl.builders.graphql_utils.handlers import NamedObjectHandler
from qctrl.builders.type_registry import TypeOverrideMixin


class ComplexArrayHandler(
    NamedObjectHandler, TypeOverrideMixin
):  # pylint:disable=too-many-ancestors
    """
    Custom handler for ComplexArray.
    """

    _type_hint = np.ndarray
    _object_name = "ComplexArray"

    def process_loaded_data(self, data: Dict[str, Any]) -> np.ndarray:
        return read_numpy_array(
            **{underscore(key): value for key, value in data.items()}
        )

    def _load_data(self, field_type: GraphQLType, data: Dict[str, Any]) -> Any:
        return self.process_loaded_data(data)


class ComplexDenseArrayHandler(
    ComplexArrayHandler
):  # pylint:disable=too-many-ancestors
    """
    Custom handler for ComplexDenseArray.
    """

    _object_name = "ComplexDenseArray"


class ComplexSparseArrayHandler(
    ComplexArrayHandler
):  # pylint:disable=too-many-ancestors
    """
    Custom handler for ComplexSparseArray.
    """

    _object_name = "ComplexSparseArray"



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring

import gzip
from typing import (
    Any,
    Dict,
    Optional,
)

import requests
from gql import Client
from gql.transport.exceptions import (
    TransportClosed,
    TransportProtocolError,
    TransportServerError,
)
from gql.transport.requests import RequestsHTTPTransport
from graphql import (
    DocumentNode,
    ExecutionResult,
    print_ast,
)
from requests.compat import json as complexjson


class QctrlRequestsHTTPTransport(RequestsHTTPTransport):
    """Custom GQL requests class to enable compression of requests."""

    def execute(  # pylint:disable=arguments-differ
        self,
        document: DocumentNode,
        variable_values: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
        timeout: Optional[str] = None,
        disable_result_cache: bool = False,
    ) -> ExecutionResult:
        """Execute GraphQL query.

        Execute the provided document AST against the configured remote server. This
        uses the requests library to perform a HTTP POST request to the remote server.

        Parameters
        ----------
        document : DocumentNode
            GraphQL query as AST Node object.
        variable_values : Optional[Dict[str, Any]]
            Dictionary of input parameters (Default value = None).
        operation_name : Optional[str]
            Name of the operation that shall be executed.
            Only required in multi-operation documents (Default value = None).
        timeout : Optional[str]
            Specifies a default timeout for requests (Default value = None).
        disable_result_cache: bool
            disable cache layer usage when retrieve result(Default value = False).


        Returns
        -------
        ExecutionResult
            The result of execution.
            `data` is the result of executing the query, `errors` is null
            if no errors occurred, and is a non-empty array if an error occurred.

        Raises
        ------
        TransportServerError
            if the status code is 400 or higher
        TransportProtocolError
            in the other cases
        TransportClosed
            if gql client session is closed.
        """

        if not self.session:
            raise TransportClosed("Transport is not connected")

        query_str = print_ast(document)
        payload: Dict[str, Any] = {"query": query_str}
        if variable_values:
            payload["variables"] = variable_values
        if operation_name:
            payload["operationName"] = operation_name

        data_key = "json" if self.use_json else "data"

        # compress the payload
        json_payload = complexjson.dumps(payload)
        gzip_payload = gzip.compress(bytes(json_payload, "utf-8"))

        if disable_result_cache is not None:
            self.headers.update({"disable_result_cache": str(disable_result_cache)})

        post_args = {
            "headers": self.headers,
            "auth": self.auth,
            "cookies": self.cookies,
            "timeout": timeout or self.default_timeout,
            "verify": self.verify,
            data_key: gzip_payload,
        }

        # Pass kwargs to requests post method
        post_args.update(self.kwargs)

        # Using the created session to perform requests
        response = self.session.request(
            self.method, self.url, **post_args  # type: ignore
        )

        try:
            response.raise_for_status()
            result = response.json()

        except requests.HTTPError as error:
            # We raise a TransportServerError if the status code is 400 or higher
            raise TransportServerError(str(error)) from error

        except Exception as exception:
            # We raise a TransportProtocolError in the other cases
            raise TransportProtocolError(
                "Server did not return a GraphQL result"
            ) from exception

        if "errors" not in result and "data" not in result:
            raise TransportProtocolError("Server did not return a GraphQL result")

        return ExecutionResult(errors=result.get("errors"), data=result.get("data"))


class QctrlGqlClient(Client):
    """
    Custom gql client class
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # force the fetch schema for validating client and client version
        with self as session:
            session.fetch_schema()



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring

BASE_URL = "https://{}.q-ctrl.com"
DEFAULT_API_ROOT = BASE_URL.format("api")
DEFAULT_OIDC_URL = BASE_URL.format("id")
OIDC_CLIENT_ID = "boulder-opal-python"
UNIVERSAL_ERROR_MESSAGE = (
    "Oops, something went wrong. Please start again or contact support@q-ctrl.com"
)

UNEXPECTED_TOKEN_ERROR = """
---------------------------------------------------------------------
An error occurred with your session authentication. Please try again.
If the issue persists, recreate your environment authentication by
running `qctrl auth` from the command line.

For non-interactive or alternative options check our help:

    $ qctrl auth --help

---------------------------------------------------------------------
"""

INVALID_SUBSCRIPTION_ERROR = """
---------------------------------------------------------------------
Your authentication succeeded, but your subscription is invalid for
this product.

Please access `https://boulder.q-ctrl.com` for more details.
---------------------------------------------------------------------
"""



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
import os
import sys
import webbrowser
from threading import Thread
from typing import (
    Dict,
    Optional,
    Union,
)

import click
import tenacity
from gql import (
    Client,
    gql,
)
from qctrlcommons.exceptions import QctrlException
from requests.exceptions import (
    BaseHTTPError,
    RequestException,
)

from .builders import (
    AsyncResult,
    build_and_bind_namespaces,
    create_client_auth,
    create_environment,
    create_gql_client,
)
from .constants import (
    BASE_URL,
    DEFAULT_API_ROOT,
    DEFAULT_OIDC_URL,
    INVALID_SUBSCRIPTION_ERROR,
    UNEXPECTED_TOKEN_ERROR,
)
from .graphs import Graph
from .parallel import ParallelExecutionCollector
from .queries import (
    ActivityMonitorQuery,
    GetMutationNameQuery,
    MeQuery,
)
from .utils import (
    PackageRegistry,
    _get_mutation_result_type,
)

LOGGER = logging.getLogger(__name__)


class Qctrl:
    """A mediator class. Used to authenticate with Q-CTRL and access Q-CTRL features.

    Creating an instance of this class requires authentication with Q-CTRL's API.

    The recommended method of authentication is through the interactive authentication
    method. This method can be invoked by simply calling Qctrl() without any arguments.
    This method will also create an authentication file that will be used for subsequent
    authentications when using the package.

    .. code-block:: python

      qctrl = Qctrl()

    .. deprecated:: 15.4.4
          Email and password based authentication has been removed in favour of
          url based authentication. To authenticate simply call `qctrl = Qctrl()`
          without any additional arguments.

    Parameters
    ----------
    email : str, optional
        The email address for a Q-CTRL account.
        [This field has been deprecated] (Default value = None)
    password : str, optional
        The password for a Q-CTRL account. [This field has been deprecated] (Default value = None)
    api_root : str, optional
        The URL of the Q-CTRL API. (Default value = None)
    oidc_url : str, optional
        The URL of the Q-CTRL OIDC. (Default value = None)
    env_namespace : str, optional
        The environment namespace of the Q-CTRL stack. (Default value = None)
    skip_version_check: bool, optional
        Option for disabling the version check. (Default value = False)
    client : gql.Client, optional
        A GraphQL client that provides access to a Q-CTRL GraphQL endpoint. You can pass
        this parameter to use Q-CTRL features provided by a non-standard Q-CTRL API
        implementation, for example, one running locally or in a private cloud. If you
        pass this parameter, do not pass `email`, `password`, or `api_root`.
        (Default value = None)

    Attributes
    ----------
    functions : :ref:`qctrl.dynamic.namespaces.FunctionNamespace`
    types : :ref:`qctrl.dynamic.types`

    Raises
    ------
    QctrlApiException
    """

    gql_api = None
    functions = None
    types = None

    def __init__(
        self,
        email: str = None,
        password: str = None,
        api_root: str = None,
        oidc_url: str = None,
        env_namespace: Optional[str] = None,
        skip_version_check: bool = False,
        client: Client = None,
    ):
        if not skip_version_check:
            self._check_version_thread()

        if email or password:
            error_message = (
                "Q-CTRL Python package authentication has been updated."
                " Please start a session without your credentials."
            )
            click.echo(error_message)
            sys.exit(1)

        if client is None:
            api_root = api_root or os.environ.get("QCTRL_API_HOST") or DEFAULT_API_ROOT
            oidc_url = oidc_url or os.environ.get("QCTRL_OIDC_URL") or DEFAULT_OIDC_URL

            env_namespace = env_namespace or os.environ.get("QCTRL_ENV_NAMESPACE")
            if env_namespace:
                api_root = BASE_URL.format(f"{env_namespace}.api")
                oidc_url = BASE_URL.format(f"{env_namespace}.id")

            self.gql_api = _build_client(api_root=api_root, oidc_url=oidc_url)
        else:
            if env_namespace or oidc_url or api_root:
                raise ValueError(
                    "If you pass a client, do not pass an env_namespace, oidc_url, or api_root."
                )
            self.gql_api = client

        self.gql_env = create_environment(self.gql_api)
        self.collector: Optional[ParallelExecutionCollector] = None
        self._build_and_bind_namespaces()

    @tenacity.retry(
        wait=tenacity.wait_exponential(multiplier=1, min=1, max=5),
        stop=tenacity.stop_after_attempt(3),
        retry=tenacity.retry_if_exception_type((RequestException, BaseHTTPError)),
    )
    def _build_and_bind_namespaces(self):
        """Builds the dynamic namespaces."""
        build_and_bind_namespaces(self)

    @staticmethod
    def create_graph() -> Graph:
        """
        Creates a graph object for representing remote computations.

        Returns
        -------
        Graph
            The new graph object.
        """
        return Graph()

    def activity_monitor(
        self,
        limit: int = 5,
        offset: int = 0,
        status: str = None,
        action_type: str = None,
    ) -> None:
        """Prints a list of previously run actions to the console
        and their statuses. Allows users to filter the amount of
        actions shown as well as provide an offset.

        Parameters
        ----------
        limit : int
            The number of previously ran actions to show.(Default is 5)
        offset : int
            Offset the list of actions by a certain amount.
        status : str
            The status of the action.
        action_type : str
            The action type.
        """

        query = ActivityMonitorQuery(self.gql_api)
        query(limit=limit, offset=offset, status=status, action_type=action_type)

    def get_result(
        self, action_id: Union[str, int]
    ) -> "qctrl.dynamic.types.CoreActionResult":
        """This function is used to return the results of a previously run function.
        You will be able to get the id of your action from the activity monitor.

        Parameters
        ----------
        action_id: str
            the id of the action which maps to an executed function.

        Returns
        -------
        qctrl.dynamic.types.CoreActionResult
            an instance of a class derived from a CoreActionResult.
        """
        action_id = str(action_id)

        query = GetMutationNameQuery(self.gql_api)
        mutation_name = query(action_id)

        field_type = _get_mutation_result_type(self.gql_api.schema, mutation_name)

        refresh_query = self.gql_env.build_refresh_query(field_type)
        data = refresh_query(action_id)

        result = self.gql_env.load_data(field_type, data)
        self.gql_env.wait_for_completion(refresh_query, AsyncResult(field_type, result))
        return result

    def parallel(self) -> ParallelExecutionCollector:
        """
        Context manager for executing multiple function calls in parallel.

        Any :ref:`functions <qctrl.dynamic.namespaces.FunctionNamespace>` that you call inside the
        context manager will be scheduled for execution when the context manager exits. For
        example:

        .. code-block:: python

          with qctrl.parallel():
              result_1 = qctrl.functions.calculate_optimization(...)
              result_2 = qctrl.functions.calculate_optimization(...)
              # The functions get executed when the context manager exits, so result_1 and result_2
              # do not have the results of the optimizations yet.

          # Once outside the context manager, the functions have been executed, so result_1 and
          # result_2 now contain the optimization results.

        Returns
        -------
        ParallelExecutionCollector
            The context manager that collects function calls to be executed in parallel.
        """
        return ParallelExecutionCollector(self)

    def is_collecting(self) -> bool:
        """Checks if the object is in collection mode.

        Returns
        -------
        bool
            True if in collection mode, False otherwise.
        """
        return bool(self.collector)

    def start_collection_mode(self, collector: ParallelExecutionCollector):
        """Starts collection mode. All function calls will be collected
        and executed when the collector object exits.

        Parameters
        ----------
        collector: ParallelExecutionCollector
            the collector object where function calls are stored.

        Raises
        ------
        RuntimeError
            unable to enter collection mode if already collecting.
        """
        if self.is_collecting():
            raise RuntimeError("unable to nest parallel collections")

        self.collector = collector

    def stop_collection_mode(self):
        """Stops collection mode. Function calls will be executed
        immediately.
        """
        self.collector = None

    def _run_gql_query(self, query: str, variable_values: Dict = None) -> Dict:
        """
        Runs a GQL query in a Python script.

        Parameters
        ----------
        query: str
            query string.
        variable_values: Dict
            Dictionary of input parameters. (Default value = None)

        Returns
        -------
        Dict
            gql response.

        Raises
        ------
        QctrlException
            if there's any root level errors
        """
        response = self.gql_api.execute(gql(query), variable_values)
        if response.get("errors"):
            raise QctrlException(response["errors"])
        return response

    @staticmethod
    def _check_version_thread():
        """
        Use another thread to check qctrl version.
        """
        for package in PackageRegistry:
            Thread(target=package.check_latest_version, daemon=True).start()


@tenacity.retry(
    wait=tenacity.wait_exponential(multiplier=1, min=1, max=5),
    stop=tenacity.stop_after_attempt(3),
    retry=tenacity.retry_if_exception_type((RequestException, BaseHTTPError)),
)
def _build_client(api_root: str, oidc_url: str) -> Client:
    """
    Builds an authenticated GraphQL client and validates its authentication.

    Parameters
    ----------
    api_root: str
        API root (Default value = None)
    oidc_url: str
        OIDC url (Default value = None)

    Returns
    -------
    GraphQLClient
        gql client.
    """

    try:
        auth = create_client_auth(oidc_url)
        client = create_gql_client(api_root, auth)
        MeQuery(client)()

    except QctrlException as error:
        LOGGER.info(error, exc_info=error)

        if "token" in str(error):
            click.echo(UNEXPECTED_TOKEN_ERROR, err=True)
            sys.exit(1)

        if "invalid subscription" in str(error):
            if "api." in api_root:
                webbrowser.open(api_root.replace("api.", "boulder."))

            click.echo(INVALID_SUBSCRIPTION_ERROR, err=True)
            sys.exit(1)

        sys.exit(error)

    return client


__all__ = ["Qctrl"]



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
import logging
from collections import OrderedDict
from enum import (
    Enum,
    EnumMeta,
)
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

import attr
import inflection
import numpy as np
from cached_property import cached_property
from graphql import (
    GraphQLEnumType,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLList,
    GraphQLNonNull,
    GraphQLScalarType,
    GraphQLType,
    GraphQLUnionType,
)
from typing_utils import get_origin

from qctrl.builders.doc import (
    ClassAttribute,
    ClassDocstring,
)
from qctrl.dynamic import (
    BoulderModule,
    register_dynamic_class,
)
from qctrl.utils import (
    _is_deprecated,
    _is_undefined,
)

from .graphql_utils import BaseTypeRegistry
from .graphql_utils.handlers import (
    NamedInputObjectHandler,
    NamedObjectHandler,
)
from .graphql_utils.handlers.mixins import ScalarMixin

LOGGER = logging.getLogger(__name__)


class TypeOverrideMixin:  # pylint:disable=too-few-public-methods
    """Handler mixin to provide the type hint
    to be used for an overridden type.
    """

    _type_hint = None

    @property
    def type_hint(self):  # pylint:disable=missing-function-docstring
        if self._type_hint is None:
            raise RuntimeError

        return self._type_hint


class TypeRegistry(BaseTypeRegistry):
    """Builds dataclasses from GraphQL types."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._types = {}

    def has_override(self, type_name: str) -> bool:
        """
        Checks if there is an override registered for
        the given type.
        """
        return type_name in self._overrides

    def get_override(self, type_name: str) -> TypeOverrideMixin:
        """
        Returns the Override for the given type,
        assuming one exists.
        """
        if not self.has_override(type_name):
            raise KeyError(f"override does not exist: {type_name}")

        return self._overrides[type_name]

    @cached_property
    def _overrides(self):
        """Collects all type overrides from the environment."""
        _overrides = {}

        for handler in self.env.iter_handlers(TypeOverrideMixin):
            if isinstance(handler, NamedObjectHandler):
                _overrides[handler.object_name] = handler

            if isinstance(handler, NamedInputObjectHandler):
                _overrides[handler.input_object_name] = handler

        return _overrides

    @cached_property
    def _scalar_types(self):
        """Collects all scalar handlers from the environment."""
        _scalar_types = {}

        for handler in self.env.iter_handlers(ScalarMixin):
            _scalar_types[handler.scalar_name] = handler.scalar_type_hint

        return _scalar_types

    def get_scalar_type_hint(self, field_type: GraphQLScalarType) -> Optional[type]:
        """Returns the type hint for the given scalar type."""
        return self._scalar_types.get(field_type.name)

    def get_type_hint(self, field_type: GraphQLType, **build_opts) -> type:
        """Returns the type hint for the GraphQL type. If there is no
        corresponding type, a new type is created. If the type needs to be
        created, `build_opts` are passed to the `build` method.

        Parameters
        ----------
        field_type: GraphQLType
            gql field type (i.e GraphQLNonNull, GraphQLScalarType)
        **build_opts: Dict
            options for forging class.

        Returns
        -------
        type
            type hint for the field.
        """
        LOGGER.debug("getting type hint for: %s", field_type)

        # non-null
        if isinstance(field_type, GraphQLNonNull):
            return self.get_type_hint(field_type.of_type, **build_opts)

        # scalar
        if isinstance(field_type, GraphQLScalarType):
            type_hint = self.get_scalar_type_hint(field_type)

            if type_hint is None:
                LOGGER.warning("unhandled scalar type: %s", field_type)

            return type_hint

        # list
        if isinstance(field_type, GraphQLList):
            return List[self.get_type_hint(field_type.of_type, **build_opts)]

        # overridden type
        if self.has_override(field_type.name):
            override = self.get_override(field_type.name)
            return override.type_hint

        # existing type
        if self.has_type(field_type.name):
            return self.get_type(field_type.name)

        # build new type
        return self.build(field_type, **build_opts)

    def get_doc_field_type(self, field_type: GraphQLType) -> Optional[str]:
        """
        Returns the field type string to be used in a docstring.

        Parameters
        ----------
        field_type: GraphQLType
            gql field type (i.e GraphQLNonNull, GraphQLScalarType)

        Returns
        -------
        str, optional
            field type in string.
        """
        # non-null
        if isinstance(field_type, GraphQLNonNull):
            return self.get_doc_field_type(field_type.of_type)

        # scalar
        if isinstance(field_type, GraphQLScalarType):
            type_hint = self.get_scalar_type_hint(field_type)

            if type_hint is None:
                LOGGER.warning("unhandled scalar type: %s", field_type)

            else:
                if type_hint.__module__ == "builtins":
                    type_hint = type_hint.__qualname__
                else:
                    type_hint = type_hint.__module__ + "." + type_hint.__qualname__

            return type_hint

        # list
        if isinstance(field_type, GraphQLList):
            sub_type = self.get_doc_field_type(field_type.of_type)
            return f"List[{sub_type}]"

        # overridden type
        if self.has_override(field_type.name):
            override = self.get_override(field_type.name)
            return override.type_hint.__name__

        # get existing type
        if self.has_type(field_type.name):
            cls = self.get_type(field_type.name)

            # e.g qctrl.dynamic.types.filter_function.Drive
            # cls.__module = qctrl.dynamic.types.filter_function
            # cls.__name__ = Drive
            return cls.__module__ + "." + cls.__name__

        # unable to determine class name
        return None

    def has_type(self, type_name: str) -> bool:
        """
        Checks if the type has been registered.
        """
        return type_name in self._types

    def get_type(self, type_name: str) -> type:
        """
        Returns the type associated with the type name.
        """
        return self._types[type_name]

    def get_type_map(self) -> Dict[str, type]:
        """Returns the type map."""
        return self._types

    @staticmethod
    def parse_name(name: str) -> Tuple[str, List[str]]:
        """
        Parses the GraphQL object name and returns
        a tuple of the base name and the namespace.
        Example:
        Core__Special__Type => ("Type", ["Core", "Special"])
        """
        parts = name.split("__")
        return parts[-1], parts[:-1]

    def build(self, obj: GraphQLType, base: type = None) -> Union[type, Tuple[type]]:
        """
        Builds a new data class from the GraphQL type.
        """
        LOGGER.debug("building type for %s - %s", obj, type(obj))

        # handle non-null
        if isinstance(obj, GraphQLNonNull):
            return self.build(obj.of_type, base)

        # handle enum
        if isinstance(obj, GraphQLEnumType):
            # Create "attributes" for the enum values to ensure they're
            # documented.
            attrs = []
            for name, value in obj.values.items():
                attrs.append(ClassAttribute(name, description=value.description))
            docstring = ClassDocstring(description=obj.description, attrs=attrs)
            cls = self._create_enum(obj.name, obj.values, base, str(docstring))
            return cls

        if isinstance(obj, GraphQLUnionType):
            return tuple(self.build(union_type, base) for union_type in obj.types)

        # handle complex type with sub fields
        fields = OrderedDict()
        docstring = ClassDocstring(description=obj.description)

        # for each sub field
        for field_name, field in obj.fields.items():

            is_optional = False

            if _is_deprecated(field):
                LOGGER.debug("%s.%s is deprecated - skipping", obj.name, field_name)
                continue

            attr_name = self.env.field_to_attr(field_name)

            # get type hint for sub field
            type_hint = self.get_type_hint(field.type)

            # build field definition
            field_def = dict(kw_only=True, type=type_hint)

            # if building non-input type, all sub-fields are
            # optional as it's possible to request partial data
            if not isinstance(obj, GraphQLInputObjectType):
                is_optional = True
                field_def.update(dict(default=None))

            # otherwise, check if sub field has default value
            elif isinstance(field, GraphQLInputField):

                # check for default value
                if not _is_undefined(field.default_value):
                    # attribute is optional if it has default value
                    is_optional = True
                    field_def.update(dict(default=field.default_value))

                # if nullable, make optional
                elif not isinstance(field.type, GraphQLNonNull):
                    # attribute is optional if it is nullable
                    is_optional = True
                    field_def.update(dict(default=None))

            # add type validator
            validator = self._get_validator(type_hint, is_optional)
            field_def.update(dict(validator=validator))

            # add field definition
            fields[attr_name] = attr.ib(**field_def)

            # add attr for docstring
            docstring.attrs.append(
                ClassAttribute(
                    attr_name,
                    attr_type=self.get_doc_field_type(field.type),
                    description=field.description,
                    optional=is_optional,
                )
            )

        # create new attr class
        cls = self._create_attr_class(obj.name, fields, base, str(docstring))
        return cls

    @staticmethod
    def _get_validator(type_hint, is_optional):
        # get origin type from type hints if exists (i.e.: list for List[])
        base_type = get_origin(type_hint) or type_hint
        type_checking = [base_type]

        # accepts dict as custom types input
        if attr.has(base_type):
            type_checking.append(dict)

        # handle optional attributes
        if is_optional:
            type_checking.append(type(None))

        # accept ndarrays where list type is accepted
        if list in type_checking:
            type_checking.append(np.ndarray)
            type_checking.append(tuple)

        # accept float values where complex type are accepted
        if complex in type_checking:
            type_checking.append(float)
            type_checking.append(np.complexfloating)

        # accept int values where float type are accepted
        if float in type_checking:
            type_checking.append(int)
            type_checking.append(np.floating)

        if int in type_checking:
            type_checking.append(np.integer)

        # Enum values can be set as string
        if isinstance(type_hint, EnumMeta):
            type_checking.append(str)

        validator = attr.validators.instance_of(tuple(type_checking))
        return validator

    @staticmethod
    def _get_submodule(namespace: List[str]) -> str:
        submodule = "types"

        if namespace:
            submodule = f"{submodule}.{'.'.join(map(inflection.underscore, namespace))}"

        return submodule

    def _create_enum(
        self,
        object_name: str,
        values: Dict[str, Any],
        base: type = None,
        docstring: str = None,
    ) -> type:
        """
        Creates a new enum class and registers it.
        """
        name, namespace = self.parse_name(object_name)
        cls = Enum(name, values, type=base)

        if docstring:
            cls.__doc__ = str(docstring)

        self._register(cls, name=object_name, submodule=self._get_submodule(namespace))

        return cls

    def _create_attr_class(
        self,
        object_name: str,
        fields: OrderedDict,
        base: type = None,
        docstring: str = None,
    ) -> type:
        """
        Creates a new attr class and registers it.
        """
        name, namespace = self.parse_name(object_name)

        if base:
            cls = attr.make_class(name, fields, bases=(base,))
        else:
            cls = attr.make_class(name, fields)

        if docstring:
            cls.__doc__ = str(docstring)

        self._register(cls, name=object_name, submodule=self._get_submodule(namespace))

        return cls

    def _register(self, cls: type, name: str = None, submodule: str = None) -> None:
        """
        Registers the new type class.

        Raises
        -------
        KeyError
            duplicate type name.
        """
        key = name or cls.__name__

        if key in self._types:
            raise KeyError(f"duplicate type name: {key}")

        self._types[key] = cls
        register_dynamic_class(cls, submodule, BoulderModule.QCTRL)



# Copyright 2022 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
# pylint:disable=missing-module-docstring
from abc import ABC

from graphql import GraphQLType

from qctrl.utils import abstract_property


class BaseHandler(ABC):  # pylint:disable=too-few-public-methods
    """Base class for a GraphQL type handler."""

    _field_type = None

    def __init__(self, env: "GraphQLEnvironment"):
        self._env = env

    @abstract_property
    def field_type(self):  # pylint:disable=missing-function-docstring
        return self._field_type

    def _is_expected_type(self, field_type: GraphQLType) -> bool:
        """Checks if the given field type is of
        the expected type.
        """
        return isinstance(  # pylint:disable=isinstance-second-argument-not-valid-type
            field_type, self.field_type
        )
